import bk7084 as bk
from bk7084.math import Vec3

win = bk.Window()
win.set_title("BK7084 - 0x00: Window")
win.set_size(800, 600)
win.set_resizable(True)

app = bk.App()
print(type(Vec3(1, 2, 3)))
camera = app.create_camera(pos=Vec3(5, 5, 5), look_at=Vec3(0, 0, 0), fov_v=60.0)

counter = 0


@app.event
def on_update(input, dt, t):
    global counter
    counter += dt
    if input.is_key_pressed(bk.KeyCode.Space):
        print("Space key is pressed")
    if input.is_mouse_pressed(bk.MouseButton.Left):
        print("Left mouse button is pressed")
    if input.is_shift_pressed():
        print("Shift key is pressed")


@app.event
def on_resize(width, height):
    print("Window resized to %dx%d" % (width, height))


app.run(win)



import bk7084 as bk
from bk7084.math import *
from bk7084 import res_path

window = bk.Window()
window.set_title('BK7084 - Solar System')
window.set_size(1024, 1024)
window.set_resizable(True)

app = bk.App()

camera = app.create_camera(Vec3(-500.0, 50.0, 0.0), Vec3(0, 0, 0), 60.0, near=1.0, far=2000.0)

camera.set_as_main_camera()

planets = {
    'neptune': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/neptune.obj'))),
        'to_sun': 450.0,
    },
    'uranus': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/uranus.obj'))),
        'to_sun': 400.0,
    },
    'saturn': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/saturn.obj'))),
        'to_sun': 350.0,
    },
    'jupiter': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/jupiter.obj'))),
        'to_sun': 300.0,
    },
    'mars': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/mars.obj'))),
        'to_sun': 400.0,
    },
    'earth': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/earth.obj'))),
        'to_sun': 200.0,
    },
    'moon': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/moon.obj'))),
        'to_earth': 80.0,
    },
    'venus': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/venus.obj'))),
        'to_sun': 120.0,
    },
    'mercury': {
        'model': app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/mercury.obj'))),
        'to_sun': 80.0,
    },
}

sun = app.add_mesh(bk.Mesh.load_from(res_path('models/solar_system/sun.obj')))
sun.set_visible(True)
sun.set_transform(Mat4.from_scale(Vec3(0.8)))

mercury = planets['mercury']['model']
# mercury.set_visible(True)
mercury.set_transform(Mat4.from_scale(Vec3(0.8)))

venus = planets['venus']['model']
venus.set_visible(True)

earth = planets['earth']['model']
earth.set_visible(True)

moon = planets['moon']['model']
moon.set_visible(True)

mars = planets['mars']['model']
# mars.set_visible(True)
mars.set_transform(Mat4.from_scale(Vec3(2.5)))

start = 0

@app.event
def on_update(input, dt, t):
    global start
    start += dt

    t = t * 0.5

    # neptune.apply_transform(neptune_translation).then(Mat4.from_rotation_y(time * 35.0, True))

    # uranus.reset_transform()
    # uranus.apply_transform(uranus_translation).then(Mat4.from_rotation_y(time * 30.0, True))
    # draw(uranus)

    earth_transform = Mat4.from_rotation_y(t) * \
                      Mat4.from_translation(Vec3(planets['earth']['to_sun'], 0, 0)) * \
                      Mat4.from_rotation_y(t) * Mat4.from_scale(Vec3(0.3, 0.3, 0.3))

    moon_transform = earth_transform * \
                      Mat4.from_rotation_y(t * 4.0) * \
                      Mat4.from_translation(Vec3(planets['moon']['to_earth'], 0, 0)) * \
                      Mat4.from_rotation_y(t)

    venus_transform = Mat4.from_rotation_y(t * 1.5) * \
                      Mat4.from_translation(Vec3(planets['venus']['to_sun'], 0, 0)) * \
                      Mat4.from_rotation_y(t * 1.5) * Mat4.from_scale(Vec3(2.5))
    #
    # mercury.reset_transform()
    # mercury.apply_transform(Mat4.from_rotation_y(time * 4.0))\
    #     .then(Mat4.from_translation(Vec3(planets['mercury']['to_sun'], 0, 0)))\
    #     .then(Mat4.from_rotation_y(time * 4.0))
    #
    # mars.reset_transform()
    # mars.apply_transform(Mat4.from_rotation_y(time * 0.5))\
    #     .then(Mat4.from_translation(Vec3(planets['mars']['to_sun'], 0, 0)))\
    #     .then(Mat4.from_rotation_y(time * 0.5))

    earth.set_transform(earth_transform)
    moon.set_transform(moon_transform)
    venus.set_transform(venus_transform)


app.run(window)



from framework.bk7084 import Window, app
from framework.bk7084.math import Vec3, Mat4
from framework.bk7084.misc import PaletteDefault as Palette

# Setup window and add camera
from framework.bk7084.scene import Mesh, Building, Component, Scene
from framework.bk7084.scene.mesh import SubMesh

window = Window("BK7084: Construction", width=1024, height=1024)
# window.create_camera(Vec3(4, 2.0, 4.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True)


class Wall(Component):
    def __init__(self, w, h):
        super().__init__()
        self._mesh = Mesh(
            name='wall',
            vertices=[[-w / 2.0, -h / 2.0, 0.0], [w / 2.0, -h / 2.0, 0.0], [w / 2.0, h / 2.0, 0.0], [-w / 2.0, h / 2.0, 0.0],
                      [-w, -h, 0.0], [w, -h, 0.0], [w, h, 0.0], [-w, h, 0.0]],
            colors=[Palette.BlueA.as_color()],
            normals=[[0.0, 0.0, 1.0]],
            uvs=[[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0], [0.25, 0.25], [0.75, 0.25], [0.75, 0.75], [0.25, 0.75]],
            faces=[
                [(0, 1, 5, 4), (4, 5, 1, 0), (0, 0, 0, 0)],
                [(1, 2, 6, 5), (5, 6, 2, 1), (0, 0, 0, 0)],
                [(2, 3, 7, 6), (6, 7, 3, 2), (0, 0, 0, 0)],
                [(3, 0, 4, 7), (7, 4, 0, 3), (0, 0, 0, 0)],
                [(0, 1, 2, 3), (0, 1, 2, 3), (0, 0, 0, 0)]],
            vertex_shader='shaders/example.vert',
            pixel_shader='shaders/example.frag')

        self._mesh.update_sub_mesh(0, SubMesh(name='body', faces=[0, 1, 2, 3], enable_normal_map=True),
                                   texture='models/brick.jpg', normal_map='models/brick_normal_map.png')
        self._mesh.append_sub_mesh(SubMesh(name='window', faces=[4]), texture='models/window.jpg')
                                   # pixel_shader='shaders/example2.frag')
        self._mesh.texture_enabled = True
        # self._mesh.apply_transformation(Mat4.from_rotation_y(45, True))

    @property
    def mesh(self) -> Mesh:
        return self._mesh


wall = Wall(1.0, 1.0)
print(wall.mesh.sub_meshes)
print(wall.mesh.rendering_info)
# wall2 = Wall(1.0, 1.0, texture1='textures/checker_huge.png', texture2='textures/checker_large.png')
# wall2.transform = Mat4.from_translation(Vec3(1.0, 0.0, 0.0))
# wall3 = Wall(1.0, 1.0, texture1='textures/checker_small.png', texture2='textures/checker_color.png')
# wall3.transform = Mat4.from_rotation_y(90, True) * Mat4.from_translation(Vec3(1.4, 0.0, 0.0))

building = Building()
building.append(wall)
# building.append(wall2, wall)
# building.append(wall3, wall)

scene = Scene(window, [building])
scene.create_camera(Vec3(2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)


@window.event
def on_draw(dt):
    # building.draw()
    # scene.draw()
    scene.draw_v2()


@window.event
def on_key_press(key, mods):
    pass


@window.event
def on_update(dt):
    pass
    # wall.transform *= Mat4.from_rotation_y(45.0 * dt, True)
    # wall2.mesh.transformation *= Mat4.from_rotation_z(45.0 * dt, True)


app.init(window)
app.run()



__all__ = ['App', 'Window', 'LightType', 'Color', 'IllumModel']


import enum
import inspect
from bk7084.bkfw import Window, Color, IllumModel
from bk7084.bkfw import PyAppState
from bk7084.bkfw import run_main_loop

from bk7084.math import Vec3, Mat4
from bk7084.mesh import Material, Mesh


class LightType(enum.Enum):
    Directional = 0
    Point = 1


class App(PyAppState):
    """The main application class."""
    def __new__(cls):
        return super().__new__(cls)

    def event(self, *args):
        """Decorator for attaching event handlers to the window.

        Usage:

        @app.event
        def on_update(dt, input):
            pass

        @app.event('on_resize')
        def random_name(width, height):
            pass
        """
        if len(args) == 0:  # @window.event()
            def decorator(fn):
                self.attach_event_handler(fn.__name__, fn)
                return fn
            return decorator
        elif inspect.isroutine(args[0]):  # @window.event
            fn = args[0]
            self.attach_event_handler(fn.__name__, fn)
            return args[0]
        elif type(args[0]) in (str,):  # @window.event('on_resize')
            def decorator(fn):
                self.attach_event_handler(args[0], fn)
            return decorator

    def add_point_light(self, pos: Vec3, color: Color, show_light: bool = False):
        """Adds a point light to the scene.

        Args:
            pos (Vec3): Position of the light.
            color (Color): Color of the light.
            show_light (bool, optional): Whether to show the light object. Defaults to False.
        """
        light = super().add_point_light_py(pos, color)
        if show_light:
            mat = Material()
            mat.diffuse = color
            mat.illum_model = IllumModel.DiffuseNoShading
            mesh = Mesh.create_sphere(0.1, 16, 16)
            mesh.set_material(mat)
            sphere = self.add_mesh(mesh)
            sphere.set_visible(True)
            sphere.set_transform(Mat4.from_translation(pos))
            return light, sphere
        return light

    def run(self, builder: Window):
        """Starts the main loop."""
        run_main_loop(self, builder)



"""
Shader005:

Shader example using bk7084 framework, shadow mapping.

OpenGL allows us to disable writing to the depth buffer by setting its depth mask to GL_FALSE.
"""

import imgui
import numpy as np

from bk7084 import Window, app, Camera, gl, ShaderProgram, VertexShader, PixelShader
from bk7084.app import ui
from bk7084.app.input import KeyCode
from bk7084.assets import default_asset_mgr
from bk7084.graphics import draw, VertexArrayObject, VertexBuffer, VertexLayout, VertexAttrib, VertexAttribFormat, \
    DirectionalLight
from bk7084.graphics.framebuffer import Framebuffer, ColorAttachment
from bk7084.graphics.texture import Texture
from bk7084.math import Vec3, Mat4
from bk7084.misc import PaletteDefault

vertex_src = """
#version 330

layout (location = 0) in vec3 a_position;
layout (location = 1) in vec4 a_color;
layout (location = 2) in vec2 a_texcoord;
layout (location = 3) in vec3 a_normal;
layout (location = 4) in vec3 a_tangent;

out vec4 v_color;
out vec3 v_normal;
out vec2 v_texcoord;
out vec3 v_tangent;
out vec3 frag_pos;
out vec3 light_pos;
out vec3 light_dir;
out vec4 frag_pos_light_space;

uniform mat4 model_mat;
uniform mat4 view_mat;
uniform mat4 proj_mat;
uniform mat4 light_mat;
uniform vec3 in_light_pos;
uniform vec3 in_light_dir;

void main() {
    vec4 pos = view_mat * model_mat * vec4(a_position, 1.0);

    frag_pos = pos.xyz;  // vertex position in camera space
     
    frag_pos_light_space = light_mat * model_mat * vec4(a_position, 1.0);

    light_pos = vec3(view_mat * vec4(in_light_pos, 1.0));
    light_dir = vec3(view_mat * vec4(in_light_dir, 0.0));

    v_color = a_color;
    v_texcoord = a_texcoord;
    v_normal = mat3(transpose(inverse(view_mat * model_mat))) * a_normal;
    v_tangent = mat3(transpose(inverse(view_mat * model_mat))) * a_tangent;

    gl_Position = proj_mat * view_mat * model_mat * vec4(a_position, 1.0);
}
"""

fragment_src = """
# version 330

struct Material {
    sampler2D diffuse_map;
    sampler2D bump_map;
    sampler2D normal_map;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    float shininess;
    bool enabled;
    bool use_diffuse_map;
    bool use_normal_map;
    bool use_bump_map;
    bool use_parallax_map;
};


in vec4 v_color;
in vec3 v_normal;
in vec2 v_texcoord;
in vec3 v_tangent;
in vec3 frag_pos;
in vec3 world_pos;
in vec3 light_pos;
in vec3 light_dir;
in vec4 frag_pos_light_space;

out vec4 frag_color;

uniform bool shading_enabled;
uniform Material mtl;
uniform vec3 light_color;
uniform bool is_directional;

uniform bool shadow_map_enabled;
uniform sampler2D shadow_map;

float calc_shadow(vec4 pos_in_light_space, vec3 normal, vec3 l) {
    float bias = max(0.00005 * (1.0 - dot(normal, l)), 0.0005); 
    if (shadow_map_enabled) {
        vec3 proj_coords = pos_in_light_space.xyz / pos_in_light_space.w;
        proj_coords = proj_coords * 0.5 + 0.5;
        float depth = proj_coords.z;
        float max_depth = texture(shadow_map, proj_coords.xy).r; 
        return depth - bias > max_depth ? 1.0 : 0.0;
        //return depth > max_depth ? 1.0 : 0.0;
    } else {
        return 0.0;
    }
}

// Blinn Phong BRDF in Camera Space
vec3 blinnPhongBRDF(vec3 light_dir, vec3 view_dir, vec3 normal, vec3 diffuse_color, vec3 specular_color, float shininess) {
    vec3 color = diffuse_color;
    vec3 half_dir = normalize(view_dir + light_dir);
    float spec_dot = max(dot(half_dir, normal), 0.0);
    color += pow(spec_dot, shininess) * specular_color;
    return color;
}

vec4 shading(vec3 ambient_color, vec3 light_dir, vec3 view_dir, vec3 light_color, vec3 n, vec3 diffuse_color, vec3 specular_color, float shininess, float shadow) {
    vec3 luminance = ambient_color.rgb * 0.1 + diffuse_color * 0.3;

    float illuminance = dot(light_dir, n);

    if (illuminance > 0.0) {
        vec3 brdf = blinnPhongBRDF(light_dir, view_dir, n, diffuse_color.rgb, specular_color.rgb, shininess);

        luminance += (1.0 - shadow) * brdf * illuminance * light_color.rgb * 0.6;
    }

    return vec4(luminance, 1.0);
}

mat3 tangentSpaceMatrix () {

    vec3 tangent = normalize(v_tangent);
    vec3 normal = normalize(v_normal);
    vec3 bitangent = normalize (cross(normal, tangent));

    return mat3(tangent, bitangent, normal);
}

vec2 parallaxMap (sampler2D tex) { 
    float parallax_factor = 10.0;

    vec3 view_dir = normalize(-frag_pos);

    float height =  texture(tex, v_texcoord).x;
    vec2 displacement = view_dir.xy / view_dir.z * (height * parallax_factor);

    return v_texcoord - displacement;
} 

vec3 bumpMap (sampler2D tex) {

    //https://developer.download.nvidia.com/CgTutorial/cg_tutorial_chapter08.html
    vec2 tex_size = 1.0/textureSize(tex, 0);


    // texture gradient (only x component is necessary since bump map is gray scale)
    vec2 grad = vec2(texture(tex, v_texcoord).x - texture(tex, v_texcoord+vec2(tex_size.x,0)).x,
                  texture(tex, v_texcoord+vec2(0,tex_size.y)).x - texture(tex, v_texcoord).x);

    // bump map multiplier (to enhance effect)
    grad *= 3.0;

    // cross product of vector (0, 1, grad.x) x (1, 0, grad.y)
    vec3 bump_map = normalize(vec3(-grad.y, grad.x, 1.0));

    // place bump_map in tangent space
    mat3 TBN = tangentSpaceMatrix ();
    bump_map = normalize(TBN * bump_map);

    return bump_map;

}

vec3 normalMap (sampler2D tex) {

    mat3 TBN = tangentSpaceMatrix ();

    vec3 normal_map = texture(tex, v_texcoord).xyz*2.0 - vec3(1.0);
    normal_map = normalize(TBN * normal_map);

    return normal_map;
}

void main() {
    vec3 _light_dir = vec3(0.0);
    if (is_directional)
        _light_dir = -normalize(light_dir);
    else
        _light_dir = normalize(light_pos - frag_pos);
    
    vec3 view_dir = normalize(-frag_pos);
    vec3 n = normalize(v_normal);

    vec4 diffuse_color;
    vec4 specular_color;
    float shininess;
    vec4 ambient_color;

    if (mtl.enabled) {
        if (mtl.use_diffuse_map) {
            diffuse_color = texture(mtl.diffuse_map, v_texcoord);
        } else {
            diffuse_color = vec4(mtl.diffuse, 1.0);
        }
    
        if (mtl.use_normal_map) {
            n = normalMap(mtl.normal_map);
        }

        else if (mtl.use_bump_map) {
            n = bumpMap(mtl.diffuse_map);
            diffuse_color = vec4(mtl.diffuse, 1.0);
        }
        else if (mtl.use_parallax_map) {
            vec2 tex_displacement = parallaxMap(mtl.diffuse_map);
            diffuse_color = texture(mtl.diffuse_map, tex_displacement);
        }

        specular_color = vec4(mtl.specular, 1.0);
        shininess = mtl.shininess;
        ambient_color = vec4(mtl.ambient, 1.0);
    } else {
        diffuse_color = v_color;
        specular_color = vec4(1.0, 1.0, 1.0, 1.0);
        shininess = 1.0;
        ambient_color = vec4(0.5, 0.5, 0.5, 1.0);
    }

    float shadow = calc_shadow(frag_pos_light_space, n, _light_dir);

    if (shading_enabled) {
        frag_color = shading(ambient_color.xyz, _light_dir, view_dir.xyz, light_color.xyz, n, diffuse_color.xyz, specular_color.xyz, shininess, shadow);
    } else {
        frag_color = diffuse_color;
    }
}
"""


# Setup window and add camera
from bk7084.scene import Mesh, Scene

window = Window("BK7084: shadow mapping", width=1024, height=1024)

sphere = Mesh("./models/uv_sphere.obj",
              texture='./textures/checker.png',
              texture_enabled=True
              # vertex_shader=vertex_src, pixel_shader=fragment_src
              )
# cube.cast_shadow = True
# cube.shading_enabled = False
# cube.texture_enabled = False
# print(sphere.alternate_texture_enabled)
cube = default_asset_mgr.get_or_create_mesh_data('cube', 'models/cube.obj')
cube.texture_enabled = True

model = Mesh("./models/spot_cow.obj",
             texture='./textures/checker.png',
             # vertex_shader=vertex_src, pixel_shader=fragment_src
             )
model.cast_shadow = True
model.apply_transform(Mat4.from_translation(Vec3(2.0, 0.0, 2.0)))
model.texture_enabled = True

cube.transformation = model.transformation

ground = Mesh(
    vertices=[[-10.0, 0.0, -10.0],
              [10.0,  0.0, -10.0],
              [10.0,  0.0,  10.0],
              [-10.0, 0.0,  10.0]],
    colors=[PaletteDefault.GreenB.as_color()],
    normals=[[0, 1, 0]],
    uvs=[[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0]],
    faces=[[(0, 1, 2, 3), (0, 1, 2, 3), (0, 0, 0, 0)]],
    texture='./textures/checker.png',
    # vertex_shader=vertex_src,
    # pixel_shader=fragment_src,
)
ground.texture_enabled = True

# scene = Scene(window, [cube, ground], light=DirectionalLight(), draw_light=False)
scene = Scene(window, [sphere, ground, model, cube], draw_light=True)
scene.create_camera(Vec3(6, 6.0, 6.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False, near=0.1, far=100.0)

animate = False

# framebuffer = Framebuffer(1024, 1024, depth_shader_accessible=True)
# depth_map_shader = ShaderProgram(
#     VertexShader.from_file('./shaders/depth_map.vert'),
#     PixelShader.from_file('./shaders/depth_map.frag')
# )


@window.event
def on_draw(dt):
    # # first pass: render to depth map
    # with framebuffer:
    #     framebuffer.enable_depth_test()
    #     framebuffer.clear(PaletteDefault.BlueB.as_color().rgba)
    #     scene.draw(shader=depth_map_shader)
    #
    # # second pass: render scene as normal with shadow mapping (using depth map from 1st pass)
    # gl.glClearColor(*PaletteDefault.Background.as_color().rgba)
    # gl.glEnable(gl.GL_DEPTH_TEST)
    # gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    # scene.draw(shadow_map=framebuffer.depth_attachments[0], shadow_map_enabled=True)

    # scene.draw(auto_shadow=False)
    scene.draw_v2(auto_shadow=True)

    # # first pass: render to depth map
    # with scene._framebuffer:
    #     scene._framebuffer.enable_depth_test()
    #     scene._framebuffer.clear(PaletteDefault.BlueB.as_color().rgba)
    #     scene.draw(shader=scene._depth_map_pipeline)
    #
    # # second pass: render scene as normal with shadow mapping (using depth map from 1st pass)
    # gl.glClearColor(*PaletteDefault.Background.as_color().rgba)
    # gl.glEnable(gl.GL_DEPTH_TEST)
    # gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    # scene.draw(shadow_map=scene._framebuffer.depth_attachments[0], shadow_map_enabled=True)


@window.event
def on_key_press(key, mods):
    global animate
    if key == KeyCode.A:
        animate = not animate

    if key == KeyCode.C:
        scene._depth_map_framebuffer.save_color_attachment()

    if key == KeyCode.D:
        scene._depth_map_framebuffer.save_depth_attachment(scene.main_camera.near, scene.main_camera.far, False)

    if key == KeyCode.Up:
        ground.apply_transform(Mat4.from_translation(Vec3(0.0, 0.1, 0.0)))

    if key == KeyCode.Down:
        ground.apply_transform(Mat4.from_translation(Vec3(0.0, -0.1, 0.0)))


@window.event
def on_update(dt):
    if animate:
        sphere.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), 45 * dt, True))
        model.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), -45 * dt, True))
        cube.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), -15 * dt, True))


app.init(window)
app.run()



import argparse
import os.path
import sys

import bk7084 as bk
from bk7084.math import *


def read_from_file(filepath, scale=1.0):
    if not os.path.exists(filepath):
        print("File not found: %s" % filepath)
        return
    transform = Mat4.from_scale(Vec3(scale, scale, scale))
    if os.path.isdir(filepath):
        for root, dirs, files in os.walk(filepath):
            for filename in files:
                if filename.endswith(".obj"):
                    model = app.add_mesh(bk.Mesh.load_from(os.path.join(root, filename)))
                    model.set_visible(True)
                    model.set_transform(transform)
    else:
        model = app.add_mesh(bk.Mesh.load_from(filepath))
        model.set_visible(True)
        model.set_transform(transform)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="*", help="Files to open")
    parser.add_argument("--far", type=int, default=200, help="Far plane")
    parser.add_argument("--cam-pos", type=float, nargs=3, default=[5, 5, 5], help="Camera position")
    parser.add_argument("--scale", type=float, default=1.0, help="Scale")

    args = parser.parse_args()

    win = bk.Window()
    win.set_title("BK7084 - Viewer")
    win.set_size(800, 600)
    win.set_resizable(True)

    app = bk.App()
    camera = app.create_camera(pos=Vec3(args.cam_pos), look_at=Vec3(0, 0, 0), fov_v=60.0, near=0.1, far=args.far)
    camera.set_as_main_camera()

    app.add_directional_light(Vec3(-1.0, -1.0, -1.0), bk.Color.WHITE)
    app.add_directional_light(Vec3(1.0, 1.0, 1.0), bk.Color.WHITE)
    # app.add_point_light(Vec3(2.5, 0, 2.5), bk.Color(1.0, 1.0, 1.0), show_light=True)

    for filepath in args.files:
        read_from_file(filepath, args.scale)

    @app.event
    def on_update(input, dt, t):
        pass


    app.run(win)



__all__ = ['Ray', 'Triangle']

from bk7084.math import Vec3


class Ray:
    def __init__(self, origin: Vec3, direction: Vec3):
        self._origin = origin
        self._direction = direction

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, o: Vec3):
        self._origin = o

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, d: Vec3):
        self._direction = d


class Triangle:
    def __init__(self, p0: Vec3, p1: Vec3, p2: Vec3):
        self._p0 = p0
        self._p1 = p1
        self._p2 = p2

    @property
    def p0(self):
        return self._p0

    @property
    def p1(self):
        return self._p1

    @property
    def p2(self):
        return self._p2


import tkinter.filedialog

import imgui
from bk7084 import Window, app, Camera
from bk7084 import ui, KeyCode
from bk7084.graphics import draw
from bk7084.math import Vec3, Mat4

# Setup window and add camera
from framework.bk7084.scene import Mesh, Scene

window = Window("BK7084: Simple Scene", width=600, height=600)

scene = Scene(window, draw_light=True)
model = scene.load_mesh_entity("spot_cow", "./models/spot_cow.obj")

scene.create_camera(Vec3(2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)
scene.create_camera(Vec3(-2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)
scene.create_camera(Vec3(2, 1.0, -2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)

animate = False


@window.event
def on_draw(dt):
    # scene.draw()
    scene.draw_v2()


@window.event
def on_key_press(key, mods):
    global animate
    if key == KeyCode.A:
        animate = not animate


@window.event
def on_update(dt):
    if animate:
        pass
        model.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), 45.0 * dt, True))


@window.event
def on_gui():
    global model
    if ui.tree_node('Mesh'):
        if ui.radio_button('texture', model.meshes[0][0].texture_enabled):
            model.meshes[0][0].texture_enabled = not model.meshes[0][0].texture_enabled

        if ui.radio_button('normal map', model.meshes[0][0].normal_map_enabled):
            model.meshes[0][0].normal_map_enabled = not model.meshes[0][0].normal_map_enabled

        if ui.radio_button('bump map', model.meshes[0][0].bump_map_enabled):
            model.meshes[0][0].bump_map_enabled = not model.meshes[0][0].bump_map_enabled

        if ui.radio_button('parallax map', model.meshes[0][0].parallax_map_enabled):
            model.meshes[0][0].parallax_map_enabled = not model.meshes[0][0].parallax_map_enabled
        ui.tree_pop()

    if ui.button("Open"):
        file = tkinter.filedialog.askopenfilename(filetypes=[("OBJ", "*.obj")])
        if file:
            model = scene.load_mesh_entity(file)


app.init(window)
app.run()



import math
import numbers

import numpy as np

from .quaternion import Quat
from .vector import Vec2, Vec3, Vec4, cross


class Matrix(np.ndarray):
    """Base class for Mat3 and Mat4. Data is stored in column major.

    Numpy stores the matrix by default in row-major order.
    """
    _shape = None

    def __new__(cls, *args, **kwargs):
        dtype = kwargs.get('dtype', np.float32)

        if len(args) == 0:
            return np.zeros(cls._shape, dtype=dtype).view(dtype)

        elif len(args) == 1:
            if isinstance(args[0], (list, tuple, np.ndarray)):
                obj = np.array(args[0], dtype=dtype)
                obj = obj.reshape(cls._shape)
            elif isinstance(args[0], numbers.Number):
                obj = np.zeros(cls._shape, dtype=dtype)
                np.fill_diagonal(obj, args[0])
            else:
                raise ValueError("Non number objects can't be matrix's elements")

            if obj.ndim > 2:
                raise ValueError("Matrix must be 2-dimensional")

            return obj.view(cls)

        else:
            obj = np.array([*args], dtype=dtype)

            return obj.view(cls)

    def __array_finalize__(self, obj):
        if self._shape != obj.shape:
            raise ValueError(f"{type(self).__name__} only accepts 1-D array of size {self._shape[0] * self._shape[1]}"
                             f" or 2-D array of size {self._shape[1]}x{self._shape[1]}")

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot add a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            return np.add(self, other)

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot subtract a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            return np.subtract(self, other)

    def __iadd__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot add a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            self[:] = np.add(self, other)
            return self

    def __isub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot add a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            self[:] = np.subtract(self, other)
            return self

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            return np.multiply(self, other).view(type(self))
        elif isinstance(other, np.ndarray) and self.shape[1] == other.shape[0]:
            result = np.matmul(self, other)
            if isinstance(other, Vec2):
                return result.view(Vec2)
            elif isinstance(other, Vec3):
                return result.view(Vec3)
            elif isinstance(other, Vec4):
                return result.view(Vec4)
            elif isinstance(other, Matrix):
                return result.view(type(self))
            else:
                return result.view(np.ndarray)
        else:
            raise TypeError(f"Cannot multiply a '{self.__class__.__name__}' '{type(other).__name__}'")

    def __rmul__(self, other):
        if isinstance(other, numbers.Number):
            print('yes')
            return np.multiply(other, self).view(type(self))
        elif isinstance(other, np.ndarray) and other.shape[1] == self.shape[0]:
            return np.matmul(other, self)
        else:
            raise TypeError(f"Cannot multiply a '{type(other).__name__}' by '{self.__class__.__name__}'")

    def __imul__(self, other):
        self[:] = self.__mul__(other)
        return self

    def __idiv__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError(f"Cannot divide a '{self.__class__.__name__}' by '{type(other).__name__}'")
        else:
            self[:] = np.divide(self, other)
            return self

    def __eq__(self, other):
        return np.asarray(self) == np.asarray(other)

    def col(self, i):
        """Returns the matrix column for the given index `i`."""
        return self[:, i]

    def row(self, i):
        """Returns the matrix row for the given index `i`."""
        return self[i, :]

    @property
    def determinant(self):
        """Returns the determinate of the matrix."""
        return 0

    @property
    def inverse(self):
        """Returns the (multiplicative) inverse of invertible `self`."""
        return np.linalg.inv(self)

    @property
    def transpose(self):
        """Returns the transpose of the matrix.
        """
        return super(Matrix, self).transpose(self)

    @property
    def conjugate_transpose(self):
        return self.transpose.conjugate()

    @classmethod
    def identity(cls, dtype=np.float32):
        return np.identity(cls._shape[0], dtype=dtype).view(cls)

    @classmethod
    def from_mat2(cls, mat: np.ndarray, dtype=None):
        """Creates an affine transformation matrix from the given 2x2 matrix."""
        if mat.shape != (2, 2):
            raise ValueError("Input matrix doesn't have shape of 2 x 2.")
        intype = dtype
        if dtype is None:
            intype = np.float32 if mat.dtype is None else mat.dtype
        m = cls.identity(intype)
        m[:2, :2] = mat[:, :]
        return m.view(cls)

    @classmethod
    def from_mat3(cls, mat: np.ndarray, dtype=None):
        """Creates a 2x2 matrix from a 3x3 matrix, discarding the last row and column."""
        if mat.shape != (3, 3):
            raise ValueError("Input matrix doesn't have shape of 3 x 3.")
        intype = dtype
        if dtype is None:
            intype = np.float32 if mat.dtype is None else mat.dtype
        m = cls.identity(intype)
        if cls._shape[0] <= 3:
            m[:, :] = mat[:cls._shape[0], :cls._shape[1]]
        else:
            m[:3, :3] = mat[:, :]
        return m.view(cls)

    @classmethod
    def from_mat4(cls, mat: np.ndarray, dtype=None):
        """Creates an affine transformation matrix from the given 4x4 matrix."""
        if mat.shape != (4, 4):
            raise ValueError("Input matrix doesn't have shape of 4 x 4.")
        intype = dtype
        if dtype is None:
            intype = np.float32 if mat.dtype is None else mat.dtype
        m = cls.identity(intype)
        m[:, :] = m[:cls._shape[0], :cls._shape[1]]
        return m.view(cls)

    @classmethod
    def from_diagonal(cls, diagonal: np.ndarray, dtype=np.float32):
        """Creates a matrix from a vector."""
        m = np.zeros(cls._shape, dtype=dtype)
        np.fill_diagonal(m, diagonal)
        return m.view(cls)

    def apply(self, vec):
        """Apply the rotation onto a vector."""
        pass


class Mat2(Matrix):
    _shape = (2, 2)

    def __new__(cls, *args, **kwargs):
        return super(Mat2, cls).__new__(cls, *args, **kwargs)

    def row(self, i):
        return Vec2(self[i, :])

    def col(self, i):
        return Vec2(self[:, i])

    @classmethod
    def from_rows(cls, x_axis, y_axis):
        """Creates a 2x2 matrix from two row vectors."""
        return cls([x_axis, y_axis])

    @classmethod
    def from_cols(cls, x_axis, y_axis):
        """Creates a 2x2 matrix from two column vectors."""
        return cls([[x_axis[0], y_axis[0]], [x_axis[1], y_axis[1]]])

    @classmethod
    def from_rotation(cls, angle: np.float32, dtype=np.float32):
        """Creates a 2x2 matrix containing a rotation of `angle` (in radians)."""
        cos = math.cos(angle)
        sin = math.sin(angle)
        return cls([[cos, -sin],
                    [sin, cos]], dtype=dtype)

    def apply(self, vec):
        if not isinstance(vec, Vec2):
            raise ValueError('Mat2 can only be applied to 2-D vector.')
        return (self * vec).view(Vec2)


class Mat3(Matrix):
    _shape = (3, 3)

    def __new__(cls, *args, **kwargs):
        return super(Mat3, cls).__new__(cls, *args, **kwargs)

    def row(self, i):
        return Vec3(self[i, :])

    def col(self, i):
        return Vec3(self[:, i])

    @classmethod
    def from_cols(cls, x_axis: Vec3, y_axis: Vec3, z_axis: Vec3, dtype=np.float32):
        """Creates a 3x3 matrix from three column vectors."""
        return cls([[x_axis[0], y_axis[0], z_axis[0]],
                    [x_axis[1], y_axis[1], z_axis[1]],
                    [x_axis[2], y_axis[2], z_axis[2]]], dtype=dtype)

    @classmethod
    def from_rows(cls, x_axis: Vec3, y_axis: Vec3, z_axis: Vec3, dtype=np.float32):
        """Creates a 3x3 matrix from three row vectors."""
        return cls([x_axis, y_axis, z_axis], dtype=dtype)

    @classmethod
    def from_euler_angles(cls, seq, angles, degrees=False, dtype=np.float32):
        """Creates a Matrix from the specified Euler angles (more precisely, Tait-Bryan angles, in radians).

        Rotations in 3-D can be represented by a sequence of 3 rotations around sequence of axes.

        Args:
            seq (str):
                Specifies sequence of axes for rotations. Up to 3 characters belonging to the set {'x', 'y', 'z'}.

            angles (np.float32 or array_like, shape (N,) or list):
                Euler angles specified in radians (`degrees` is False) or in degrees (`degrees` is True)
                For a single character in `seq`, `angles` can be:
                  - a single value
                For 2- and 3-character wide `seq`, `angles` can be:
                  - array_like with shape (N,) where `N` is the width of `seq`
                  - list of values with minimal size of `N`, where `N` is the width of `seq`

            degrees (bool):
                If True, then the given angles are assumed to be in degrees. Default is False.

            dtype (dtype of numpy):
                Specifies that element type inside of the created matrix.

        Returns:
            Mat3

        Examples:
            Initialize a single rotation matrix along a single axis:

            >>> m = Mat3.from_euler_angles('x', 90, degrees=True)

            Initialize a single rotation matrix with a give axis sequence:

            >>> m = Mat3.from_euler_angles('yzx', [30, 60, 90], degrees=True)
        """
        if seq is None or len(seq) == 0:
            raise ValueError('Not a valid sequence.')
        angles = list([angles]) if not isinstance(angles, (tuple, list, np.ndarray)) else angles
        if len(angles) != len(seq):
            raise ValueError('Provided values don\'t match the length of rotation sequence.')
        if seq.count('x') > 1 or seq.count('y') > 1 or seq.count('z') > 1:
            raise ValueError('Rotation axis can only appear one time in sequence.')
        m = cls.identity(dtype)
        for axis, angle in zip(seq, angles):
            if axis == 'x':
                m = cls.from_rotation_x(angle, degrees, dtype) * m
            if axis == 'y':
                m = cls.from_rotation_y(angle, degrees, dtype) * m
            if axis == 'z':
                m = cls.from_rotation_z(angle, degrees, dtype) * m
        return m

    @classmethod
    def from_axis_angle(cls, axis: Vec3, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from a normalized rotation `axis` and `angle` (in radians)."""
        if axis == Vec3.unit_x(dtype):
            return cls.from_rotation_x(angle, degrees, dtype)
        elif axis == Vec3.unit_y(dtype):
            return cls.from_rotation_y(angle, degrees, dtype)
        elif axis == Vec3.unit_z(dtype):
            return cls.from_rotation_z(angle, degrees, dtype)
        else:
            cos = math.cos(angle)
            sin = math.sin(angle)
            a = 1. - cos
            ux, uy, uz = axis.x, axis.y, axis.z
            uxx, uxy, uxz = ux * ux, ux * uy, ux * uz
            uyz, uyy, uzz = uy * uz, uy * uy, uz * uz
            return cls([[cos + uxx * a, uxy * a - uz * sin, uxz * a + uy * sin],
                        [uxy * a + uz * sin, cos + uyy * a, uyz * a - ux * sin],
                        [uxz * a - uy * sin, uyz * a + ux * sin, cos + uzz * a]], dtype=dtype)

    @classmethod
    def from_rotation_x(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the x axis."""
        angle = angle if not degrees else math.radians(angle)
        cos = math.cos(angle)
        sin = math.sin(angle)
        return cls([[1., 0., 0.],
                    [0., cos, -sin],
                    [0., sin, cos]], dtype=dtype)

    @classmethod
    def from_rotation_y(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the y axis."""
        angle = angle if not degrees else math.radians(angle)
        cos = math.cos(angle)
        sin = math.sin(angle)
        return cls([[cos, 0., sin],
                    [0., 1., 0.],
                    [-sin, 0., cos]], dtype=dtype)

    @classmethod
    def from_rotation_z(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the z axis."""
        angle = angle if not degrees else math.radians(angle)
        cos = math.cos(angle)
        sin = math.sin(angle)
        return cls([[cos, -sin, 0.],
                    [sin, cos, 0.],
                    [0., 0., 1.]], dtype=dtype)

    @classmethod
    def from_scale(cls, scale: Vec3, dtype=np.float32):
        """Creates an affine transformation matrix from the given non-uniform 3D `scale`.
        The resulting matrix can be used to transform 3D points and vectors."""
        return cls.from_diagonal(scale, dtype)

    @classmethod
    def from_translation(cls, translation: Vec2, dtype=np.float32):
        """Creates an affine transformation matrix from the given 2D `translation`.
        The resulting matrix can be used to transform 2D points and vectors."""
        m = cls.identity(dtype=dtype)
        m[:2, 2] = translation
        return m

    @classmethod
    def from_mat4(cls, mat: np.ndarray, dtype=None):
        """Creates an affine transformation matrix from the given 4x4 matrix.
        The resulting matrix can be used to transform 2D points and vectors."""
        assert (mat.shape == (4, 4))
        intype = dtype
        if dtype is None:
            intype = np.float32 if mat.dtype is None else mat.dtype
        m = np.zeros(cls._shape, dtype=intype)
        m[:3, :3] = mat[:3, :3]
        return m

    @classmethod
    def from_quat(cls, quat: Quat, dtype=np.float32):
        """Creates an affine transformation matrix from the given quaternion (unit, normalised)."""
        xx, yy, zz = quat.x * quat.x, quat.y * quat.y, quat.z * quat.z
        xy, xz, xw = quat.x * quat.y, quat.x * quat.z, quat.x * quat.w
        yz, yw, zw = quat.y * quat.z, quat.y * quat.w, quat.z * quat.w
        return cls([[1. - 2. * yy - 2. * zz, 2. * xy - 2. * zw, 2. * xz + 2. * yw],
                    [2. * xy + 2. * zw, 1. - 2. * xx - 2. * zz, 2. * yz - 2. * xw],
                    [2. * xz - 2. * yw, 2. * yz + 2. * xw, 1. - 2. * xx - 2. * yy]], dtype=dtype)

    def apply(self, vec):
        if not isinstance(vec, Vec3):
            raise ValueError('Mat3 can only be applied to 3-D vector.')
        return (self * vec).view(Vec3)


class Mat4(Matrix):
    _shape = (4, 4)

    def __new__(cls, *args, **kwargs):
        return super(Mat4, cls).__new__(cls, *args, **kwargs)

    def row(self, i):
        return Vec4(self[i, :])

    def col(self, i):
        return Vec4(self[:, i])

    @classmethod
    def from_rows(cls, x_axis: Vec4, y_axis: Vec4, z_axis: Vec4, w_axis: Vec4):
        """Creates a 4x4 matrix from four column vectors."""
        return cls([x_axis, y_axis, z_axis, w_axis])

    @classmethod
    def from_cols(cls, x_axis: Vec4, y_axis: Vec4, z_axis: Vec4, w_axis: Vec4):
        """Creates a 3x3 matrix from three column vectors."""
        return cls([[x_axis[0], y_axis[0], z_axis[0], w_axis[0]],
                    [x_axis[1], y_axis[1], z_axis[1], w_axis[1]],
                    [x_axis[2], y_axis[2], z_axis[2], w_axis[2]],
                    [x_axis[3], y_axis[3], z_axis[3], w_axis[3]]])

    @classmethod
    def from_euler_angles(cls, seq, angles, degrees=False, dtype=np.float32):
        """Creates a Matrix from the specified Euler angles (more precisely, Tait-Bryan angles, in radians).

        Rotations in 3-D can be represented by a sequence of 3 rotations around sequence of axes.

        Args:
            seq (str):
                Specifies sequence of axes for rotations. Up to 3 characters belonging to the set {'x', 'y', 'z'}.

            angles (np.float32 or array_like, shape (N,) or list):
                Euler angles specified in radians (`degrees` is False) or in degrees (`degrees` is True)
                For a single character in `seq`, `angles` can be:
                    - a single value
                For 2- and 3-character wide `seq`, `angles` can be:
                    - array_like with shape (N,) where `N` is the width of `seq`
                    - list of values with minimal size of `N`, where `N` is the width of `seq`

            degrees (bool):
                If True, then the given angles are assumed to be in degrees. Default is False.

            dtype (dtype of numpy):
                Specifies that element type inside of the created matrix.

            Returns:
                Mat3

            Examples:
                Initialize a single rotation matrix along a single axis:

                >>> m = Mat3.from_euler_angles('x', 90, degrees=True)

                Initialize a single rotation matrix with a give axis sequence:

                >>> m = Mat3.from_euler_angles('yzx', [30, 60, 90], degrees=True)
        """
        return cls.from_mat3(Mat3.from_euler_angles(seq, angles, degrees, dtype), dtype).view(cls)

    @classmethod
    def from_axis_angle(cls, axis: Vec3, angle, degrees=False, dtype=np.float32):
        """Creates an affine transformation matrix containing a rotation around a normalized
        rotation `axis` and `angle` (in radians)."""
        return cls.from_mat3(Mat3.from_axis_angle(axis, angle, degrees, dtype), dtype)

    @classmethod
    def from_rotation_x(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the x axis."""
        return cls.from_mat3(Mat3.from_rotation_x(angle, degrees, dtype), dtype)

    @classmethod
    def from_rotation_y(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the y axis."""
        return cls.from_mat3(Mat3.from_rotation_y(angle, degrees, dtype), dtype)

    @classmethod
    def from_rotation_z(cls, angle, degrees=False, dtype=np.float32):
        """Creates a 3D rotation matrix from `angle` (in radians) around the z axis."""
        return cls.from_mat3(Mat3.from_rotation_z(angle, degrees, dtype), dtype)

    @classmethod
    def from_scale(cls, scale: Vec3, dtype=np.float32):
        """Creates an affine transformation matrix from the given non-uniform 2D `scale`.
        The resulting matrix can be used to transform 2D points and vectors."""
        return cls.from_diagonal(Vec4.from_vec3(scale, 1.), dtype=dtype)

    @classmethod
    def from_translation(cls, translation: Vec3, dtype=np.float32):
        """Creates an affine transformation matrix from the given 3D `translation`."""
        m = cls.identity(dtype=dtype)
        m[:3, 3] = np.asarray(translation).reshape(3, )
        return m

    @classmethod
    def from_quat(cls, quat, dtype=np.float32):
        """Creates an affine transformation matrix from the given quaternion (normalised)."""
        return cls.from_mat3(Mat3.from_quat(quat, dtype=dtype))

    @classmethod
    def look_at_gl(cls, eye: Vec3, center: Vec3, up: Vec3, dtype=np.float32):
        """Creates a right-handed view matrix using a camera position, and up direction, and a focal point.
        This is the same as the OpenGL `gluLookAt` function. See
        https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/gluLookAt.xml
        """
        return cls.look_at_rh_y_up(eye, center, up, dtype)

    @classmethod
    def look_at_rh_y_up(cls, eye: Vec3, center: Vec3, up: Vec3, dtype=np.float32):
        forward = (center - eye).normalised
        up = up.normalised
        u = cross(forward, up).normalised
        v = cross(u, forward).normalised
        w = -forward
        return Mat4.from_rows(Vec4(u, -u.dot(eye)),
                              Vec4(v, -v.dot(eye)),
                              Vec4(w, -w.dot(eye)),
                              Vec4(0., 0., 0., 1.))

    @classmethod
    def perspective_gl(cls, fov_y, aspect_ratio, z_near, z_far, degrees=False, dtype=np.float32):
        """Creates a right-handed perspective projection matrix with [-1, 1] depth range for a symmetric
        perspective-view frustum.

        The source coordinate space is right-handed and y-up, the destination space is left-handed
        and y-up, with Z(depth) clip extending from -1.0 (close) to 1.0 (far).

        This is the same as the OpenGL `gluPerspective` function. See
        https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/gluPerspective.xml

        Args:
            fov_y (np.float32):
                Specifies the field of view angle in the y-direction, in radians if `degrees` not set to True.

            aspect_ratio (np.float32):
                Specifies the aspect ratio that determines the field of view in the x-direction. The aspect ratio is the
                ratio of x (width) to y (height).

            z_near (np.float32):
                Specifies the near distance from the viewer to the near clipping plane (not in the world frame,
                always positive).

            z_far (np.float32):
                Specifies the far distance from the viewer to the far clipping plane (not in the world frame,
                always positive).

            degrees (bool):
                If True, then the given field of view angle is assumed to be in degrees. Default is False.

            dtype (dtype of numpy):
                Specifies that element type inside of the created matrix.

        Returns:
            Mat4
        """
        fov = math.radians(fov_y) if degrees else fov_y
        n, f = z_near, z_far
        tan = math.tan(fov / 2.)
        return cls([[1. / (tan * aspect_ratio), 0., 0., 0.],
                    [0., 1. / tan, 0., 0.],
                    [0., 0., -(f + n) / (f - n), -2. * f * n / (f - n)],
                    [0., 0., -1., 0.]], dtype=dtype)

    @classmethod
    def orthographic_gl(cls, left, right, bottom, top, near, far, dtype=np.float32):
        """Creates a right-handed orthographic projection matrix with [-1, 1] depth range.

        The source coordinate space is right-handed and y-up, the destination space is left-handed
        and y-up, with Z(depth) clip extending from -1.0 (close) to 1.0 (far).

        This is the same as the OpenGL `glOrtho` function. See
        https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/glOrtho.xml"""
        _a = 1. / (right - left)
        _b = 1. / (top - bottom)
        _c = 1. / (far - near)
        return cls([[2. * _a, 0., 0., -(right + left) * _a],
                    [0., 2. * _b, 0., -(top + bottom) * _b],
                    [0., 0., -2. * _c, -(far + near) * _c],
                    [0., 0., 0., 1.]], dtype=dtype)

    def apply(self, vec):
        if not isinstance(vec, Vec4):
            raise ValueError('Mat4 can only be applied to 4-D vector.')
        return (self * vec).view(Vec4)



from __future__ import annotations

import math
import numbers

import numpy as np

from .vector import Vec3

"""
TODO:
  - power of a quaternion
  - dot product
  - cross product
  - create from rotation matrix
  - into rotation matrix
  - lerp
  - slerp
  - power
  - log
  - test
"""


class Quat(np.ndarray):
    _shape = (4,)

    def __new__(cls, x, y, z, w=1., dtype=float):
        return np.array([w, x, y, z]).astype(dtype).view(cls)

    def __array_finalize__(self, obj):
        pass

    def __str__(self):
        _str = f'Quat [' + '{:>11.8f},' * 3 + ' {:>11.8f}]'
        return _str.format(*self.flatten())

    def __repr__(self):
        return super(Quat, self).__repr__()

    def __len__(self):
        return 4

    def __neg__(self):
        return np.multiply(self, -1)

    def __add__(self, other):
        if not isinstance(other, Quat):
            raise ValueError('Quaternion can only add with another Quaternion')
        else:
            return np.add(self, other)

    def __sub__(self, other):
        if not isinstance(other, Quat):
            raise ValueError('Quaternion can only be subtracted by another Quaternion')
        else:
            return np.subtract(self, other)

    def __mul__(self, other):
        """Hamilton product of two quaternions.
        (w0, v0) * (w1, v1) = (w0w1 - v0*v1, w0v1 + w1v0 + v0 x v1)
        """
        if isinstance(other, numbers.Number):
            return np.multiply(self, other)
        elif isinstance(other, Quat):
            if self.is_real and other.is_real:
                return Quat(self[0] * other[0], 0., 0., 0.)
            elif self.is_pure and other.is_pure:
                v0 = self[1:]
                v1 = other[1:]
                _dot = np.dot(v0, v1)
                _cross = np.cross(v0, v1)
                return Quat(_cross[0], _cross[1], _cross[2], -_dot)
            else:
                w0, v0 = self[0], self[1:]
                w1, v1 = other[0], other[1:]
                val = w0 * v1 + w1 * v0 + np.cross(v0, v1)
                return Quat(val[0], val[1], val[2], self.w * other.w - np.dot(v0, v1))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        if isinstance(other, numbers.Number) and other != 0:
            return np.divide(self, other)
        else:
            raise ValueError('Quaternion can only be divided by a non-zero number')

    def __iadd__(self, other):
        self[:] = self.__add__(other)
        return self

    def __isub__(self, other):
        self[:] = self.__sub__(other)
        return self

    def __imul__(self, other):
        self[:] = self.__mul__(other)
        return self

    def __idiv__(self, other):
        self[:] = self.__div__(other)
        return self

    @classmethod
    def identity(cls) -> Quat:
        return Quat(0., 0., 0., 1., float)

    @classmethod
    def from_rotation_x(cls, angle, degrees=False, dtype=float):
        angle = math.radians(angle) if degrees else angle
        half = angle * 0.5
        return Quat(math.sin(half), 0., 0., math.cos(half), dtype=dtype)

    @classmethod
    def from_rotation_y(cls, angle, degrees=False, dtype=float):
        angle = math.radians(angle) if degrees else angle
        half = angle * 0.5
        return Quat(0., math.sin(half), 0., math.cos(half), dtype=dtype)

    @classmethod
    def from_rotation_z(cls, angle, degrees=False, dtype=float):
        angle = math.radians(angle) if degrees else angle
        half = angle * 0.5
        return Quat(0., 0., math.sin(half), math.cos(half), dtype=dtype)

    @classmethod
    def from_euler_angles(cls, angle, degrees=False, dtype=float):
        pass

    @classmethod
    def from_axis_angle(cls, axis, angle, degrees=False, dtype=float):
        dtype = axis.dtype or dtype
        angle = math.radians(angle) if degrees else angle
        if isinstance(axis, Vec3):
            axis.normalise()
        elif isinstance(axis, np.ndarray) and axis.shape == (3,):
            if not np.isclose(np.linalg.norm(axis), 1.):
                axis.view(Vec3).normalise()
        elif isinstance(axis, (list, tuple)) and len(axis) == 3:
            axis = np.asarray(axis)
            if not np.isclose(np.linalg.norm(axis), 1.):
                axis.view(Vec3).normalise()
        else:
            raise ValueError('Axis is not in array like form')

        half = angle * 0.5
        v = axis * math.sin(half)
        return Quat(v.x, v.y, v.z, math.cos(half))

    @classmethod
    def from_matrix(cls, mat, dtype=float):
        pass

    @property
    def conjugate(self):
        return Quat(-self.x, -self.y, -self.z, self.w)

    @property
    def norm(self):
        """The norm of the quaternion equals to the square root of the product of a quaternion with its
         conjugate is called its norm."""
        return math.sqrt(np.sum(self ** 2))

    @property
    def inverse(self):
        if self.is_nonzero:
            return self.conjugate / np.sum(self ** 2)
        else:
            raise ZeroDivisionError("Zero quaternion doesn't have inverse")

    @property
    def normalised(self):
        """Returns this quaternion with a magnitude of 1"""
        if not self.is_nonzero:
            return self
        else:
            return self[:] / self.norm

    def normalise(self):
        self[:] = self.normalised()
        return self

    @property
    def is_nonzero(self) -> bool:
        return self.any()

    @property
    def is_real(self) -> bool:
        """A Real quaternion is a quaternion with a vector term of 0."""
        return not self[1:].any()

    @property
    def is_pure(self) -> bool:
        """A Pure quaternion has a zero scalar term."""
        return self.w == 0

    @property
    def is_unit(self) -> bool:
        """A unit quaternion has a zero scalar and a unit vector."""
        return self.is_pure and np.sum(self[1:] ** 2) == 1

    @property
    def w(self):
        return self[0]

    @w.setter
    def w(self, value):
        self[0] = value

    @property
    def x(self):
        return self[1]

    @x.setter
    def x(self, value):
        self[1] = value

    @property
    def y(self):
        return self[2]

    @y.setter
    def y(self, value):
        self[2] = value

    @property
    def z(self):
        return self[3]

    @z.setter
    def z(self, value):
        self[3] = value

    @property
    def rotation_angle(self):
        """Calculates the rotation angle around the quaternion's axis.
        Returns:
            float, rotation angle in radians
        """
        return math.acos(self.w) * 2.0

    @property
    def rotation_axis(self):
        sin = math.sin(math.acos(self.w))
        return Vec3(self[1:]) / sin

    def lerp(self, other, t):
        """Interpolates between two quaternions by `t`. `t` is clamped to the range [0, 1]"""
        return NotImplementedError

    def slerp(self, other, t):
        """Spherically interpolates between two quaternions by `t`. `t` is clamped to the range [0, 1]"""
        return NotImplementedError

    def squad(self):
        """Spherical and Quadrangle, it smoothly interpolates over a path of rotations."""
        pass

    def dot(self, other):
        return NotImplementedError

    def cross(self, other):
        """Returns the cross-product of the two quaternions.
        Quaternions are NOT communicative. Therefore, order is important.

        Note:
            This is NOT the same as a vector cross-product. Quaternion cross-product is the equivalent
            of matrix multiplication.
        """
        return NotImplementedError

    def distance(self, other: Quat):
        """Computes the angular difference between two quaternions."""
        return self.inverse * other

    def exp(self):
        """Computes the exponential of the quaternion"""
        e = math.exp(self.w)
        vec_norm = np.linalg.norm(self[1:])
        if np.isclose(vec_norm, 0):
            return Quat(0, 0, 0, e, dtype=self.dtype)

        sin = np.sin(vec_norm) / vec_norm
        return Quat(self.x * sin, self.y * sin, self.z * sin, np.cos(vec_norm), dtype=self.dtype) * e

    def log(self):
        pass

    def to_mat3(self):
        return NotImplementedError

    def to_mat4(self):
        return NotImplementedError


class DualQuat(np.ndarray):
    _shape = (8,)

    def __new__(cls, real, dual):
        pass



import bk7084 as bk
from bk7084.math import *

win = bk.Window()
win.set_title("BK7084 - 0x01: Camera")
win.set_size(800, 600)
win.set_resizable(True)

app = bk.App()

cameras = [
    app.create_camera(pos=Vec3(5, 5, 5), look_at=Vec3(0, 0, 0), fov_v=60.0),
    app.create_camera(pos=Vec3(0, 0, 5), look_at=Vec3(0, 0, 0), fov_v=60.0),
    app.create_camera(pos=Vec3(0, 5, 5), look_at=Vec3(0, 0, 0), fov_v=60.0)
]

cameras[0].set_as_main_camera()

tri_mesh = bk.Mesh.create_triangle(Vec3(-2, -2, -1.5), Vec3(2, -2, -1.5), Vec3(0, 2, -1.5))
tri = app.add_mesh(tri_mesh)
tri.set_visible(True)

cube_mesh = bk.Mesh.load_from(bk.res_path("../data/blender_cube/cube.obj"))
cube = app.add_mesh(cube_mesh)
cube.set_visible(True)

i = 0
space_pressed = False

@app.event
def on_update(input, dt, t):
    global i
    global space_pressed

    if input.is_key_pressed(bk.KeyCode.Space):
        if space_pressed is False:
            space_pressed = not space_pressed
            i += 1
            cameras[i % 3].set_as_main_camera()

    if input.is_key_released(bk.KeyCode.Space):
        if space_pressed is True:
            space_pressed = not space_pressed


app.run(win)



import imgui
import numpy as np

from bk7084 import Window, app, Camera, gl, ShaderProgram, VertexShader, PixelShader
from bk7084.app import ui
from bk7084.app.input import KeyCode
from bk7084.assets import default_asset_mgr
from bk7084.graphics import draw, VertexArrayObject, VertexBuffer, VertexLayout, VertexAttrib, VertexAttribFormat
from bk7084.graphics.framebuffer import Framebuffer, ColorAttachment
from bk7084.graphics.texture import Texture
from bk7084.math import Vec3, Mat4
from bk7084.misc import PaletteDefault

vertex_src = """
# version 330

layout (location = 0) in vec3 a_position;
layout (location = 1) in vec2 a_texcoord;

out vec2 v_texcoord;

uniform mat4 model_mat;

void main()
{
    gl_Position = model_mat * vec4(a_position, 1.0);
    v_texcoord = a_texcoord;
}
"""

fragment_src = """
# version 330
in vec2 v_texcoord;

uniform sampler2D screen_texture;

out vec4 frag_color;

void main()
{
    frag_color = texture(screen_texture, v_texcoord);
}
"""

# Setup window and add camera
from bk7084.scene import Mesh, Scene

window = Window("BK7084: Framebuffer", width=1024, height=1024)

cube = Mesh("./models/spot_cow.obj")

scene = Scene(window, [cube], draw_light=True)
scene.create_camera(Vec3(2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)

animate = False

framebuffer = Framebuffer(1024, 1024, depth_shader_accessible=True)
screen_shader = ShaderProgram(
    VertexShader(vertex_src),
    PixelShader(fragment_src)
)
vao = VertexArrayObject()
vbo = VertexBuffer(4, VertexLayout((VertexAttrib.Position, VertexAttribFormat.Float32, 3),
                                   (VertexAttrib.TexCoord0, VertexAttribFormat.Float32, 2)))
vbo.set_data(np.array([-0.5, -0.5, 0.0, 0.0, 0.0,
                       0.5, -0.5, 0.0, 1.0, 0.0,
                       0.5, 0.5, 0.0, 1.0, 1.0,
                       -0.5, 0.5, 0.0, 0.0, 1.0], dtype=np.float32))
vao.bind_vertex_buffer(vbo, [0, 1])

model_mat = Mat4.identity()


@window.event
def on_draw(dt):
    # first pass
    with framebuffer:
        framebuffer.clear(PaletteDefault.BlueB.as_color().rgba)
        framebuffer.enable_depth_test()
        scene.draw()

    # second pass
    gl.glClearColor(*PaletteDefault.RedB.as_color().rgba)
    gl.glDisable(gl.GL_DEPTH_TEST)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    with screen_shader:
        screen_shader['model_mat'] = model_mat
        screen_shader['screen_texture'] = 0
        screen_shader.active_texture_unit(0)
        with framebuffer.color_attachments[0]:
            with vao:
                gl.glDrawArrays(gl.GL_TRIANGLE_FAN, 0, 4)


@window.event
def on_key_press(key, mods):
    global animate
    if key == KeyCode.A:
        animate = not animate

    if key == KeyCode.C:
        framebuffer.save_color_attachment()

    if key == KeyCode.D:
        framebuffer.save_depth_attachment(scene.main_camera.near, scene.main_camera.far)


@window.event
def on_update(dt):
    if animate:
        cube.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), 45 * dt, True))
        global model_mat
        model_mat = Mat4.from_rotation_z(45 * dt, True) * model_mat


app.init(window)
app.run()



import bk7084 as bk
from bk7084.math import *

win = bk.Window()
win.set_title("BK7084 - 0x00: Window")
win.set_size(800, 600)
win.set_resizable(True)

app = bk.App()
camera = app.create_camera(pos=Vec3(0, 0, 90), look_at=Vec3(0, 0, 0), fov_v=60.0, near=0.1, far=200.0)
camera.set_as_main_camera()

cube_mesh = bk.Mesh.load_from(bk.res_path("../data/blender_cube/cube.obj"))
sphere_mesh = bk.Mesh.load_from(bk.res_path("../data/blender_sphere/sphere.obj"))

app.add_directional_light(Vec3(0.0, 0.0, -1.0), bk.Color.WHITE)

objs = []

for i in range(0, 16):
    for j in range(0, 16):
        if j % 2 == 0:
            for k in range(0, 16):
                cube = app.add_mesh(cube_mesh)
                cube.set_visible(True)
                objs.append(cube)
        else:
            for k in range(0, 16):
                sphere = app.add_mesh(sphere_mesh)
                sphere.set_visible(True)
                objs.append(sphere)


@app.event
def on_update(input, dt, t):
    angle = t
    for i, c in enumerate(objs):
        axis = i % 3
        rot = Mat4.identity()
        if axis == 0:
            rot = Mat4.from_rotation_x(angle)
        elif axis == 1:
            rot = Mat4.from_rotation_y(angle)
        elif axis == 2:
            rot = Mat4.from_rotation_z(angle)

        z = ((i // 256) - 8) * 4.0
        x = (((i % 256) // 16) - 8) * 4.0
        y = (((i % 256) % 16) - 8) * 4.0

        c.set_transform(Mat4.from_translation(Vec3(x, y, z)) * rot)


app.run(win)



import imgui

from bk7084 import Window, app, Camera
from bk7084.app import ui
from bk7084.app.input import KeyCode
from bk7084.graphics import draw
from bk7084.math import Vec3, Mat4

# Setup window and add camera
from bk7084.scene import Mesh, Scene

window = Window("BK7084: Simple Scene", width=600, height=600)

model = Mesh("./models/spot_cow.obj")
scene = Scene(window, [model], draw_light=True)
scene.create_camera(Vec3(2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)
scene.create_camera(Vec3(-2, 1.0, 2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)
scene.create_camera(Vec3(2, 1.0, -2.0), Vec3(0, 0, 0), Vec3.unit_y(), 60.0, zoom_enabled=True, safe_rotations=False)

animate = False


@window.event
def on_draw(dt):
    # scene.draw()
    scene.draw_v2()


@window.event
def on_key_press(key, mods):
    global animate
    if key == KeyCode.A:
        animate = not animate


@window.event
def on_update(dt):
    if animate:
        pass
        model.apply_transform(Mat4.from_axis_angle(Vec3.unit_y(), 45.0 * dt, True))


@window.event
def on_gui():
    if ui.tree_node('Mesh'):
        if ui.radio_button('normal map', model.normal_map_enabled):
            model.normal_map_enabled = not model.normal_map_enabled

        if ui.radio_button('bump map', model.bump_map_enabled):
            model.bump_map_enabled = not model.bump_map_enabled

        if ui.radio_button('parallax map', model.parallax_map_enabled):
            model.parallax_map_enabled = not model.parallax_map_enabled
        ui.tree_pop()


app.init(window)
app.run()



from .input import *
from .app import *
from .camera import *
from .mesh import *
from . import math
from . import geometry
from bk7084.bkfw import ConcatOrder, Alignment


def res_path(path: str) -> str:
    """Returns the absolute path of a resource file.

    Args:
        path (str): Relative path of the resource file.

    Returns:
        str: Absolute path of the resource file.
    """
    import os.path as osp
    import traceback
    stack = traceback.extract_stack()
    dirname = osp.dirname(osp.abspath(stack[-2].filename))
    return osp.abspath(osp.join(dirname, path))
